//
// Pklad 3.2 - Genertor krajiny metodou pesouvn
//               prostednho bodu
//



#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <time.h>


// Globln promnn a funkce,
// kter jsou pouity v modulech Main-*.cpp
const char *appTitle = "Midpoint Terrain 1";      // Titulek okna
bool headlight = false;                           // Defaultn svtlo (v pozici kamery)
SoSeparator* createScene(int argc, char **argv);  // Funkce vytvejc scnu



// Velikost generovan krajiny zvten o jednu
// (krajina 32x32 se konstruuje ze 33x33 bod)
#define TERRAIN_DIMENSION 33
// Maximln vka a hloubka krajiny
#define TERRAIN_HEIGHT_Q 10.f



float hmap[TERRAIN_DIMENSION][TERRAIN_DIMENSION];


// Funkce pracujc s vkovou mapou
float randf(float max);
void generateMap();
SoSeparator* createModel();



// Funkce vracejc nhodnou float hodnotu v rozsahu -max a +max
float randf(float max)
{
  return (float(rand()) / float(RAND_MAX)) * max*2 - max;
}



// Tato funkce vytvo metodou pesouvn prostednho bodu
// vkovou mapu, kter je vrcena funkc.
// Parametr dimension poet dk a sloupc vkov mapy
// a maxHeight maximln vku krajiny.
void generateMap()
{
  // Pia poten vku do roh mapy
  hmap[0][0] = 0.f;
  hmap[0][TERRAIN_DIMENSION-1] = 0.f;
  hmap[TERRAIN_DIMENSION-1][0] = 0.f;
  hmap[TERRAIN_DIMENSION-1][TERRAIN_DIMENSION-1] = 0.f;

  // V nsledujc smyce se generuj postupn ni a ni detaily
  // mapy. Zan se s nejvtm krokem (promnn step nastavena na
  // velikost pole), a kon se s krokem 2, kdy jsou znmy vechny
  // body se sudou souadnic na obou osch a dopotvaj se
  // chybjc body mezi nimi.
  int step = TERRAIN_DIMENSION-1;
  float maxR = TERRAIN_HEIGHT_Q;
  int x,y;
  while (step >= 2) {

    // 1. fze: tvercov mka vodorovn
    // Ze ty bod tverce vygeneruj prostedn bod a k jeho hodnot
    // piti nhodn slo v rozsahu -maxR do +maxR
    for (y=0; y<TERRAIN_DIMENSION-1; y+=step) {
      for (x=0; x<TERRAIN_DIMENSION-1; x+=step) {
        hmap[y+step/2][x+step/2] = (hmap[y][x] + hmap[y][x+step] +
            hmap[y+step][x] + hmap[y+step][x+step]) / 4.f + randf(maxR);
      }
    }

    // Korekce nhodnho rozptylu vky
    maxR *= (float)sqrt(0.5);

    // 2. fze: tvercov mka pootoen o 45 stup
    // Ze ty bod opt vygeneruj prostedn bod a jeho hodnotu
    // modifikuj. Po dokonen tto fze je hustota platnch
    // platnch dat v na vkov map dvojnsobn.
    for (y=0; y<TERRAIN_DIMENSION-1; y+=step) {
      for (x=0; x<TERRAIN_DIMENSION-1; x+=step) {
        hmap[y][x+step/2] = ((y-step/2 >= 0 ? hmap[y-step/2][x+step/2] : 0) +
                            hmap[y][x] +
                            hmap[y][x+step] +
                            hmap[y+step/2][x+step/2]) /
                           (y-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[y+step/2][x] = (hmap[y][x] +
                            (x-step/2 >= 0 ? hmap[y+step/2][x-step/2] : 0) +
                            hmap[y+step/2][x+step/2] +
                            hmap[y+step][x]) /
                           (x-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[y+step/2][x+step] = (hmap[y][x+step] +
                                 hmap[y+step/2][x+step/2] +
                                 (x+step+step/2 < TERRAIN_DIMENSION ?
                                     hmap[y+step/2][x+step+step/2] : 0) +
                                 hmap[y+step][x+step]) /
                                (x+step+step/2 < TERRAIN_DIMENSION ? 4 : 3) +
                                randf(maxR);
        hmap[y+step][x+step/2] = (hmap[y+step/2][x+step/2] +
                                 hmap[y+step][x] +
                                 hmap[y+step][x+step] +
                                 (y+step+step/2 < TERRAIN_DIMENSION ?
                                     hmap[y+step+step/2][x+step/2] : 0)) /
                                (y+step+step/2 < TERRAIN_DIMENSION ? 4 : 3) +
                                randf(maxR);
      }
    }

    // Korekce nhodnho rozptylu vky
    maxR *= (float)sqrt(0.5);

    // Zmn krok na polovinu
    step /= 2;
  }
}



// Funkce vytvo model na vkov mapy
SoSeparator* createModel()
{
  // Vytvo koenov nod
  SoSeparator *root = new SoSeparator;

  // Materil (barva) krajiny
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(0.0f, 0.0f, 0.0f);
  material->diffuseColor.setValue(0.0f, 0.9f, 0.0f);
  root->addChild(material);

  // Vytvo objekt nesouc souadnice modelu krajiny
  SoCoordinate3 *coords = new SoCoordinate3;

  // Nejprve nastavme poet souadnic, kter budeme generovat.
  // (ablona SbSqr pot druhou mocninu)
  coords->point.setNum(SbSqr(TERRAIN_DIMENSION-1)*4);
  SbVec3f *c = coords->point.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako dva trojhelnky. Nepouvme u
  // SoQuadMesh, abychom v ptm pkladu mohli provdt urit kouzla.
  int ci = 0;
  int x,y;
  for (y=0; y<TERRAIN_DIMENSION-1; y++) {
    for (x=0; x<TERRAIN_DIMENSION-1; x++) {

      // Vku modelujeme na Y-ov souadnici, ta v Open Inventoru,
      // podobn jako v matematice, m nahoru, X-ov doprava a
      // zetov smrem ze scny ke kamee (tedy opan ne v Direct3D).
      c[ci++] = SbVec3f(float(x), hmap[y][x], float(y));
      c[ci++] = SbVec3f(float(x), hmap[y+1][x], float(y+1));
      c[ci++] = SbVec3f(float(x+1), hmap[y][x+1], float(y));
      c[ci++] = SbVec3f(float(x+1), hmap[y+1][x+1], float(y+1));
    }
  }
  coords->point.finishEditing();
  root->addChild(coords);


  // Nakonfiguruj triangleStripSet
  SoTriangleStripSet *stripSet = new SoTriangleStripSet;
  stripSet->numVertices.setNum(SbSqr(TERRAIN_DIMENSION-1));

  // Prvnmu ze strip vdy pia 5 souadnic (vytvo 3 tojhelnky)
  // a druhmu 3 (vytvo jeden trojhelnk).
  int32_t *nv = stripSet->numVertices.startEditing();
  for (y=0; y<SbSqr(TERRAIN_DIMENSION-1); y++)
    nv[y] = 4;

  stripSet->numVertices.finishEditing();
  root->addChild(stripSet);

  return root;
}



// Funkce vytvo graf scny pro nai aplikaci.
SoSeparator* createScene(int argc, char **argv)
{
  // Inicializace genertoru nhodnch sel
  srand((unsigned int)time(NULL));

  // Vytvo koen scny
  SoSeparator *root = new SoSeparator;
  root->ref();

  // Nastav intenzitu globlnho svtla ve scn
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.5);
  root->addChild(env);

  // Svtlo
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(0.35f, -1.0f, 0.25f);
  sunLight->intensity.setValue(1.0f);
  root->addChild(sunLight);

  // Model krajiny
  generateMap();
  SoSeparator *terrain = createModel();
  root->addChild(terrain);

  root->unrefNoDelete();
  return root;
}
